from abc import abstractmethod

from foxes.utils import new_instance
import foxes.variables as FV

from .model import Model
from .wake_superposition import WindVectorWakeSuperposition


class WakeModel(Model):
    """
    Abstract base class for wake models.

    :group: core

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__()
        self._has_uv = False

    @property
    def affects_ws(self):
        """
        Flag for wind speed wake models

        Returns
        -------
        dws: bool
            If True, this model affects wind speed

        """
        return False

    @property
    def affects_downwind(self):
        """
        Flag for downwind or upwind effects
        on other turbines

        Returns
        -------
        dwnd: bool
            Flag for downwind effects by this model

        """
        return True

    @property
    def has_uv(self):
        """
        This model uses wind vector data

        Returns
        -------
        hasuv: bool
            Flag for wind vector data

        """
        return self._has_uv

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if self.affects_ws and algo.wake_deflection.has_uv:
            self._has_uv = True
        super().initialize(algo, verbosity, force)

    @abstractmethod
    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_targets, n_tpoints, ...)

        """
        pass

    @abstractmethod
    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        pass

    def finalize_wake_deltas(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        wake_deltas,
    ):
        """
        Finalize the wake calculation.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        wake_deltas: dict
            The wake deltas object at the selected target
            turbines. Key: variable str, value: numpy.ndarray
            with shape (n_states, n_targets, n_tpoints)

        """
        pass

    @classmethod
    def new(cls, wmodel_type, *args, **kwargs):
        """
        Run-time wake model factory.

        Parameters
        ----------
        wmodel_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """
        return new_instance(cls, wmodel_type, *args, **kwargs)


class SingleTurbineWakeModel(WakeModel):
    """
    Abstract base class for wake models that represent
    a single turbine wake

    Single turbine wake models depend on superposition models.

    Attributes
    ----------
    wind_superposition: str
        The wind superposition model name (vector or compenent model),
        will be looked up in model book
    other_superpositions: dict
        The superpositions for other than (ws, wd) variables.
        Key: variable name str, value: The wake superposition
        model name, will be looked up in model book
    vec_superp: foxes.core.WindVectorWakeSuperposition or None
        The wind vector wake superposition model
    superp: dict
        The superposition dict, key: variable name str,
        value: `foxes.core.WakeSuperposition`

    :group: models.wake_models

    """

    def __init__(self, wind_superposition=None, other_superpositions={}):
        """
        Constructor.

        Parameters
        ----------
        wind_superposition: str, optional
            The wind superposition model name (vector or compenent model),
            will be looked up in model book
        other_superpositions: dict
            The superpositions for other than (ws, wd) variables.
            Key: variable name str, value: The wake superposition
            model name, will be looked up in model book

        """
        super().__init__()
        self.wind_superposition = wind_superposition
        self.other_superpositions = other_superpositions
        self.vec_superp = None
        self.superp = {}

        for v in [FV.WS, FV.WD]:
            assert v not in other_superpositions, (
                f"Wake model '{self.name}': Found variable '{v}' among 'other_superposition' keyword, use 'wind_superposition' instead"
            )

        self.__has_vector_superp = False

    @property
    def has_vector_wind_superp(self):
        """
        This model uses a wind vector superposition

        Returns
        -------
        hasv: bool
            Flag for wind vector superposition

        """
        return self.__has_vector_superp

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        w = [self.vec_superp] if self.vec_superp is not None else []
        return w + list(self.superp.values())

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        self.superp = {
            v: algo.mbook.wake_superpositions[s]
            for v, s in self.other_superpositions.items()
        }

        if self.wind_superposition is not None:
            self.vec_superp = algo.mbook.wake_superpositions[self.wind_superposition]
            self.__has_vector_superp = isinstance(
                self.vec_superp, WindVectorWakeSuperposition
            )
            if self.__has_vector_superp:
                self._has_uv = True
            else:
                self.superp[FV.WS] = self.vec_superp
                self.vec_superp = None

        super().initialize(algo, verbosity, force)

    def finalize_wake_deltas(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        wake_deltas,
    ):
        """
        Finalize the wake calculation.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        wake_deltas: dict
            The wake deltas object at the selected target
            turbines. Key: variable str, value: numpy.ndarray
            with shape (n_states, n_targets, n_tpoints)

        """
        for v in wake_deltas.keys():
            if v != FV.UV:
                try:
                    wake_deltas[v] = self.superp[v].calc_final_wake_delta(
                        algo, mdata, fdata, tdata, v, wake_deltas[v]
                    )
                except KeyError:
                    raise KeyError(
                        f"Wake model '{self.name}': Variable '{v}' appears to be modified, missing superposition model"
                    )

        if FV.UV in wake_deltas:
            assert self.has_vector_wind_superp, (
                f"{self.name}: Expecting wind vector superposition, got '{self.wind_superposition}'"
            )
            dws, dwd = self.vec_superp.calc_final_wake_delta_uv(
                algo, mdata, fdata, tdata, wake_deltas.pop(FV.UV)
            )

            wake_deltas[FV.WS] = dws
            wake_deltas[FV.WD] = dwd


class TurbineInductionModel(SingleTurbineWakeModel):
    """
    Abstract base class for turbine induction models.

    :group: core

    """

    @property
    def affects_downwind(self):
        """
        Flag for downwind or upwind effects
        on other turbines

        Returns
        -------
        dwnd: bool
            Flag for downwind effects by this model

        """
        return False

    @classmethod
    def new(cls, induction_type, *args, **kwargs):
        """
        Run-time turbine induction model factory.

        Parameters
        ----------
        induction_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """
        return new_instance(cls, induction_type, *args, **kwargs)


class WakeK(Model):
    """
    Handler for the wake growth parameter k

    Attributes
    ----------
    k_var: str
        The name of the k variable
    ti_var: str
        The name of the TI variable

    :group: core

    """

    def __init__(
        self,
        k=None,
        ka=None,
        kb=None,
        k_var=FV.K,
        ti_var=FV.TI,
    ):
        """
        Constructor.

        Parameters
        ----------
        k: float, optional
            The k value
        ka: float, optional
            The ka value in k = ka * TI + kb
        kb: float, optional
            The kb value in k = ka * TI + kb
        k_var: str
            The name of the k variable
        ti_var: str
            The name of the TI variable

        """
        super().__init__()
        self._k = k
        self._ka = ka
        self._kb = kb
        self.k_var = k_var
        self.ti_var = ti_var

        if k is not None and (ka is not None or kb is not None):
            raise ValueError("Got 'k' and also ('ka' or 'kb') as non-None parameters")
        elif k is None and kb is not None and (ka is None or ka == 0):
            raise ValueError(f"Got k={k}, ka={ka}, kb={kb}, use k={kb} instead")

        setattr(self, self.k_var, None)

    def repr(self):
        """
        Provides the representative string

        Returns
        -------
        s: str
            The representative string

        """
        if self._k is not None:
            s = f"{self.k_var}={self._k}"
        elif self._ka is not None or self._kb is not None:
            s = f"{self.k_var}={self._ka}*{self.ti_var}"
            if self._kb is not None and self._kb > 0:
                s += f"+{self._kb}"
        else:
            s = f"k_var={self.k_var}"
        return s

    @property
    def is_kTI(self):
        """Flag for ka != 0"""
        return self._ka is not None and self._ka != 0

    @property
    def all_none(self):
        """Flag for k=ka=kb=None"""
        return self._k is None and self._ka is None and self._kb is None

    @property
    def use_amb_ti(self):
        """Flag for using ambient ti"""
        return self.ti_var in FV.amb2var

    def __call__(
        self,
        *args,
        lookup_ti="w",
        lookup_k="sw",
        ti=None,
        amb_ti=None,
        **kwargs,
    ):
        """
        Gets the k value

        Parameters
        ----------
        args: tuple, optional
            Arguments for get_data
        lookup_ti: str
            The ti lookup order for get_data
        lookup_k: str
            The k lookup order for get_data
        ti: numpy.ndarray, optional
            ti data in the requested target shape,
            if known
        amb_ti: numpy.ndarray, optional
            Ambient ti data in the requested target shape,
            if known
        kwargs: dict, optional
            Arguments for get_data

        Returns
        -------
        k: numpy.ndarray
            The k array as returned by get_data

        """
        sel = kwargs.pop("selection", None)
        setattr(self, self.k_var, self._k)
        if self._ka is not None or self._kb is not None:
            if self.ti_var == FV.TI and ti is not None:
                pass
            elif self.ti_var == FV.AMB_TI and amb_ti is not None:
                ti = amb_ti
            else:
                ti = self.get_data(self.ti_var, *args, lookup=lookup_ti, **kwargs)
            kb = 0 if self._kb is None else self._kb
            setattr(self, self.k_var, self._ka * ti + kb)

        k = self.get_data(self.k_var, *args, lookup=lookup_k, selection=sel, **kwargs)
        setattr(self, self.k_var, None)
        return k



from .pool import PoolEngine as PoolEngine
from .multiprocess import MultiprocessEngine as MultiprocessEngine
from .numpy import NumpyEngine as NumpyEngine
from .single import SingleChunkEngine as SingleChunkEngine
from .mpi import MPIEngine as MPIEngine
from .ray import RayEngine as RayEngine
from .default import DefaultEngine as DefaultEngine

from .futures import ThreadsEngine as ThreadsEngine
from .futures import ProcessEngine as ProcessEngine

from .dask import DaskBaseEngine as DaskBaseEngine
from .dask import DaskEngine as DaskEngine
from .dask import LocalClusterEngine as LocalClusterEngine
from .dask import SlurmClusterEngine as SlurmClusterEngine



from .windio import windio_file2dict as windio_file2dict
from .windio import read_windio_file as read_windio_file
from .windio import read_windio_dict as read_windio_dict
from .windio import foxes_windio as foxes_windio

from .read_farm import read_layout as read_layout
from .read_farm import read_turbine_types as read_turbine_types
from .read_farm import read_hub_heights as read_hub_heights
from .read_farm import read_rotor_diameters as read_rotor_diameters
from .read_farm import read_farm as read_farm
from .read_farm import read_n_turbines as read_n_turbines

from .read_site import read_site as read_site
from .read_fields import read_wind_resource_field as read_wind_resource_field
from .read_attributes import read_attributes as read_attributes



from tqdm import tqdm
from xarray import Dataset

from foxes.core import Engine
import foxes.constants as FC

from .pool import _run


class NumpyEngine(Engine):
    """
    The numpy engine for foxes calculations.

    :group: engines

    """

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        return f(*args, **kwargs)

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        return future

    def map(
        self,
        func,
        inputs,
        *args,
        **kwargs,
    ):
        """
        Runs a function on a list of files

        Parameters
        ----------
        func: Callable
            Function to be called on each file,
            func(input, *args, **kwargs) -> data
        inputs: array-like
            The input data list
        args: tuple, optional
            Arguments for func
        kwargs: dict, optional
            Keyword arguments for func

        Returns
        -------
        results: list
            The list of results

        """
        return [func(input, *args, **kwargs) for input in inputs]

    def run_calculation(
        self,
        algo,
        model,
        model_data,
        farm_data=None,
        point_data=None,
        out_vars=[],
        chunk_store={},
        sel=None,
        isel=None,
        iterative=False,
        write_nc=None,
        **calc_pars,
    ):
        """
        Runs the model calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model: foxes.core.DataCalcModel
            The model that whose calculate function
            should be run
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset, optional
            The initial farm data
        point_data: xarray.Dataset, optional
            The initial point data
        out_vars: list of str, optional
            Names of the output variables
        chunk_store: foxes.utils.Dict
            The chunk store
        sel: dict, optional
            Selection of coordinate subsets
        isel: dict, optional
            Selection of coordinate subsets index values
        iterative: bool
            Flag for use within the iterative algorithm
        write_nc: dict, optional
            Parameters for writing results to netCDF files, e.g.
            {'out_dir': 'results', 'base_name': 'calc_results',
            'ret_data': False, 'split': 1000}.

            The split parameter controls how the output is split:
            - 'chunks': one file per chunk (fastest method),
            - 'input': split according to sizes of multiple states input files,
            - int: split with this many states per file,
            - None: create a single output file.

            Use ret_data = False together with non-single file writing
            to avoid constructing the full Dataset in memory.
        calc_pars: dict, optional
            Additional parameters for the model.calculate()

        Returns
        -------
        results: xarray.Dataset
            The model results

        """
        # subset selection:
        model_data, farm_data, point_data = self.select_subsets(
            model_data, farm_data, point_data, sel=sel, isel=isel
        )

        # basic checks:
        super().run_calculation(algo, model, model_data, farm_data, point_data)

        # prepare:
        n_states = model_data.sizes[FC.STATE]
        out_coords = model.output_coords()
        coords = {}
        if FC.STATE in out_coords and FC.STATE in model_data.coords:
            coords[FC.STATE] = model_data[FC.STATE].to_numpy()

        # DEBUG objec mem sizes:
        # from foxes.utils import print_mem
        # for m in [algo] + model.models:
        #    print_mem(m, pre_str="MULTIP CHECKING LARGE DATA", min_csize=9999)

        # calculate chunk sizes:
        n_targets = point_data.sizes[FC.TARGET] if point_data is not None else 0
        chunk_sizes_states, chunk_sizes_targets = self.calc_chunk_sizes(
            n_states, n_targets
        )
        n_chunks_states = len(chunk_sizes_states)
        n_chunks_targets = len(chunk_sizes_targets)
        self.print(
            f"{type(self).__name__}: Selecting n_chunks_states = {n_chunks_states}, n_chunks_targets = {n_chunks_targets}",
            level=2,
        )

        # prepare and submit chunks:
        n_chunks_all = n_chunks_states * n_chunks_targets
        self.print(f"{type(self).__name__}: Looping over {n_chunks_all} chunks")
        pbar = None
        if self.verbosity > 0 and self.has_progress_bar and n_chunks_all > 1:
            pbar = tqdm(total=n_chunks_all)
        results = {}
        i0_states = 0
        pdone = -1
        for chunki_states in range(n_chunks_states):
            i1_states = i0_states + chunk_sizes_states[chunki_states]
            i0_targets = 0
            for chunki_points in range(n_chunks_targets):
                key = (chunki_states, chunki_points)
                i1_targets = i0_targets + chunk_sizes_targets[chunki_points]

                # get this chunk's data:
                data = self.get_chunk_input_data(
                    algo=algo,
                    model_data=model_data,
                    farm_data=farm_data,
                    point_data=point_data,
                    states_i0_i1=(i0_states, i1_states),
                    targets_i0_i1=(i0_targets, i1_targets),
                    out_vars=out_vars,
                    chunki_states=chunki_states,
                    chunki_points=chunki_points,
                    n_chunks_states=n_chunks_states,
                    n_chunks_points=n_chunks_targets,
                )

                # submit model calculation:
                key = (chunki_states, chunki_points)
                results[key] = _run(
                    algo,
                    model,
                    *data,
                    iterative=iterative,
                    chunk_store=chunk_store,
                    chunk_key=key,
                    out_coords=out_coords,
                    write_nc=write_nc,
                    **calc_pars,
                )
                # chunk_store.update(results[key][1])
                del data

                i0_targets = i1_targets

                if pbar is not None:
                    pbar.update()
                elif self.verbosity > 0 and self.prints_progress:
                    pr = int(100 * (chunki_states + 1) / n_chunks_all)
                    if pr > pdone:
                        pdone = pr
                        print(
                            f"{self.name}: Completed {chunki_states + 1} of {n_chunks_all} chunks, {pdone}%"
                        )
            i0_states = i1_states

        if farm_data is None:
            farm_data = Dataset()
        goal_data = farm_data if point_data is None else point_data
        del calc_pars, farm_data, point_data

        if pbar is not None:
            pbar.close()
        elif self.verbosity > 0 and self.prints_progress:
            print(f"{type(self).__name__}: Completed all {i1_states} states\n")

        out = self.combine_results(
            algo=algo,
            futures=None,
            results=results,
            model_data=model_data,
            out_vars=out_vars,
            out_coords=out_coords,
            n_chunks_states=n_chunks_states,
            n_chunks_targets=n_chunks_targets,
            goal_data=goal_data,
            iterative=iterative,
            write_nc=write_nc,
        )

        return out



import time
import argparse
import matplotlib.pyplot as plt
import numpy as np

import foxes
import foxes.variables as FV


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--file_pattern",
        help="The search pattern for input *.nc files",
        default="data/data_*.nc",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default="centre", nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Jensen_linear_k007"],
        nargs="+",
    )
    parser.add_argument(
        "-wf", "--wake_frame", help="The wake frame choice", default="rotor_wd"
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-nt", "--n_turbines", help="The number of turbines", default=9, type=int
    )
    parser.add_argument("-e", "--engine", help="The engine", default="ProcessEngine")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-lm",
        "--load_mode",
        help="Do load mode",
        default="preload",
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    states = foxes.input.states.FieldData(
        args.file_pattern,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        load_mode=args.load_mode,
    )

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    farm = foxes.WindFarm()
    N = int(args.n_turbines**0.5)
    foxes.input.farm_layout.add_grid(
        farm,
        xy_base=np.array([500.0, 500.0]),
        step_vectors=np.array([[500.0, 0], [0, 500.0]]),
        steps=(N, N),
        turbine_models=args.tmodels + [ttype.name],
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        wake_models=args.wakes,
        rotor_model=args.rotor,
        wake_frame=args.wake_frame,
        partial_wakes=args.pwakes,
        mbook=mbook,
        verbosity=1,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()

        print("\nCalc time =", time1 - time0, "\n")

        print(farm_results)

        fr = farm_results.to_dataframe()
        print(fr[[FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]])

        o = foxes.output.SlicesData(algo, farm_results)
        ds = o.get_states_data_xy(
            z_list=[90, 100], variables=[FV.WS], resolution=50, verbosity=1
        )
        print(ds)

        if not args.nofig:
            o = foxes.output.FlowPlots2D(algo, farm_results)
            o.get_mean_fig_xy(FV.WS, resolution=10)
            plt.show()



from .convergence import ConvCrit as ConvCrit
from .convergence import ConvCritList as ConvCritList
from .convergence import ConvVarDelta as ConvVarDelta
from .convergence import DefaultConv as DefaultConv

from .farm_wakes_calc import FarmWakesCalculation as FarmWakesCalculation
from .urelax import URelax as URelax



from .dict import read_dict as read_dict
from .dict import run_dict as run_dict
from .dict import run_outputs as run_outputs
from .dict import run_obj_function as run_obj_function
from .dict import get_output_obj as get_output_obj

from .yaml import foxes_yaml as foxes_yaml
from .windio.windio import foxes_windio as foxes_windio



import numpy as np
from scipy.interpolate import interpn

from foxes.core import WakeFrame
from foxes.utils import wd2uv
from foxes.core.data import TData
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class Streamlines2D(WakeFrame):
    """
    Streamline following wakes

    Attributes
    ----------
    step: float
        The streamline step size in m
    cl_ipars: dict
        Interpolation parameters for centre line
        point interpolation

    :group: models.wake_frames

    """

    def __init__(self, step, max_length_km=20, cl_ipars={}, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        step: float
            The streamline step size in m
        max_length_km: float
            The maximal streamline length in km
        cl_ipars: dict
            Interpolation parameters for centre line
            point interpolation
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(max_length_km=max_length_km, **kwargs)
        self.step = step
        self.cl_ipars = cl_ipars

        self.DATA = self.var("DATA")
        self.STEPS = self.var("STEPS")
        self.SDAT = self.var("SDAT")

    def __repr__(self):
        return (
            f"{type(self).__name__}(step={self.step}, max_length={self.max_length_km})"
        )

    def _calc_streamlines(self, algo, mdata, fdata):
        """
        Helper function that computes all streamline data
        """
        # prepare:
        n_states = mdata.n_states
        n_turbines = mdata.n_turbines
        N = int(self.max_length_km * 1e3 / self.step)

        # calc data: x, y, z, wd
        data = np.zeros((n_states, n_turbines, N, 4), dtype=config.dtype_double)
        for i in range(N):
            # set streamline start point data (rotor centre):
            if i == 0:
                data[:, :, i, :3] = fdata[FV.TXYH]
                data[:, :, i, 3] = fdata[FV.AMB_WD]

            # compute next step:
            else:
                # calculate next point:
                xyz = data[:, :, i - 1, :3]
                n = wd2uv(data[:, :, i - 1, 3])
                data[:, :, i, :2] = xyz[:, :, :2] + self.step * n
                data[:, :, i, 2] = xyz[:, :, 2]

                # calculate next tangential vector:
                svars = algo.states.output_point_vars(algo)
                tdata = TData.from_points(
                    data[:, :, i, :3],
                    data={
                        v: np.full(
                            (n_states, n_turbines, 1), np.nan, dtype=config.dtype_double
                        )
                        for v in svars
                    },
                    dims={v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in svars},
                )
                data[:, :, i, 3] = algo.states.calculate(algo, mdata, fdata, tdata)[
                    FV.WD
                ][:, :, 0]

                sel = np.isnan(data[:, :, i, 3])
                if np.any(sel):
                    data[sel, i, 3] = data[sel, i - 1, 3]

        return data

    def get_streamline_data(self, algo, mdata, fdata):
        """
        Gets streamline data, generating it on the fly

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        data: numpy.ndarray
            The streamline data, shape:
            (n_states, n_turbines, n_steps, 4)
            with data x, y, z, wd

        """
        if self.DATA not in mdata or not np.all(
            mdata[self.DATA][:, :, 0, :3] == fdata[FV.TXYH]
        ):
            mdata[self.DATA] = self._calc_streamlines(algo, mdata, fdata)
            mdata.dims[self.DATA] = (FC.STATE, FC.TURBINE, self.STEPS, self.SDAT)

        return mdata[self.DATA]

    def _calc_coos(self, algo, mdata, fdata, targets, downwind_index):
        """
        Helper function, calculates streamline coordinates
        for given points and given turbine
        """
        # prepare:
        n_states, n_targets, n_tpoints = targets.shape[:3]
        n_points = n_targets * n_tpoints
        points = targets.reshape(n_states, n_points, 3)

        # find nearest streamline points:
        data = self.get_streamline_data(algo, mdata, fdata)[:, downwind_index]
        dists = np.linalg.norm(points[:, :, None, :2] - data[:, None, :, :2], axis=-1)
        selp = np.argmin(dists, axis=2)
        data = np.take_along_axis(data[:, None], selp[:, :, None, None], axis=2)[
            :, :, 0
        ]
        slen = self.step * selp
        del dists, selp

        # calculate coordinates:
        coos = np.full((n_states, n_points, 3), np.nan, dtype=config.dtype_double)
        coos[:, :, 2] = points[:, :, 2] - data[:, :, 2]
        delta = points[:, :, :2] - data[:, :, :2]
        nx = wd2uv(data[:, :, 3])
        projx = np.einsum("spd,spd->sp", delta, nx)
        sel = (projx > -self.step) & (projx < self.step)
        if np.any(sel):
            ny = np.stack([-nx[:, :, 1], nx[:, :, 0]], axis=2)
            coos[sel, 0] = slen[sel] + projx[sel]
            coos[sel, 1] = np.einsum("spd,spd->sp", delta, ny)[sel]

        return coos.reshape(n_states, n_targets, n_tpoints, 3)

    def calc_order(self, algo, mdata, fdata):
        """
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        # prepare:
        n_states = fdata.n_states
        n_turbines = algo.n_turbines
        tdata = TData.from_points(points=fdata[FV.TXYH])

        # calculate streamline x coordinates for turbines rotor centre points:
        # n_states, n_turbines_source, n_turbines_target
        coosx = np.zeros((n_states, n_turbines, n_turbines), dtype=config.dtype_double)
        for ti in range(n_turbines):
            coosx[:, ti, :] = self._calc_coos(
                algo, mdata, fdata, tdata[FC.TARGETS], ti
            )[:, :, 0, 0]

        # derive turbine order:
        # TODO: Remove loop over states
        order = np.zeros((n_states, n_turbines), dtype=config.dtype_int)
        for si in range(n_states):
            order[si] = np.lexsort(keys=coosx[si])

        return order

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        coos = self._calc_coos(algo, mdata, fdata, tdata[FC.TARGETS], downwind_index)

        return algo.wake_deflection.calc_deflection(
            algo, mdata, fdata, tdata, downwind_index, coos
        )

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        # get streamline points:
        n_states, n_points = x.shape
        data = self.get_streamline_data(algo, mdata, fdata)[:, downwind_index]
        spts = data[:, :, :3]
        n_spts = spts.shape[1]
        xs = self.step * np.arange(n_spts)

        # interpolate to x of interest:
        qts = np.zeros((n_states, n_points, 2), dtype=config.dtype_double)
        qts[:, :, 0] = np.arange(n_states)[:, None]
        qts[:, :, 1] = np.minimum(x, xs[-1])
        qts = qts.reshape(n_states * n_points, 2)
        ipars = dict(bounds_error=False, fill_value=0.0)
        ipars.update(self.cl_ipars)
        results = interpn((np.arange(n_states), xs), spts, qts, **ipars)

        return results.reshape(n_states, n_points, 3)



import numpy as np
import time
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n_s", help="The number of states", type=int, default=30)
    parser.add_argument("-n_t", help="The number of turbines", type=int, default=5)
    parser.add_argument("--n_p", help="The number of turbines", type=int, default=2000)
    parser.add_argument("--ws0", help="The lowest wind speed", type=float, default=3.0)
    parser.add_argument(
        "--ws1", help="The highest wind speed", type=float, default=30.0
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=["kTI_04"], nargs="+"
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-cl", "--calc_cline", help="Calculate centreline", action="store_true"
    )
    parser.add_argument(
        "-cm", "--calc_mean", help="Calculate states mean", action="store_true"
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    n_s = args.n_s
    n_t = args.n_t
    n_p = args.n_p
    p0 = np.array([0.0, 0.0])
    stp = np.array([500.0, 0.0])

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype
    D = ttype.D
    H = ttype.H

    states = foxes.input.states.ScanStates(
        {
            FV.WS: np.linspace(args.ws0, args.ws1, n_s),
            FV.WD: [270],
            FV.TI: [0.08],
            FV.RHO: [1.225],
        }
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_row(
        farm=farm,
        xy_base=p0,
        xy_step=stp,
        n_turbines=n_t,
        turbine_models=args.tmodels + [ttype.name],
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()

        print("\nCalc time =", time1 - time0, "\n")

        print("\nFarm results:\n", farm_results)

        o = foxes.output.FarmResultsEval(farm_results)
        o.add_capacity(algo)
        o.add_capacity(algo, ambient=True)
        o.add_efficiency()

        # state-turbine results
        farm_df = farm_results.to_dataframe()
        print("\nFarm results data:\n")
        print(
            farm_df[
                [
                    FV.X,
                    FV.WD,
                    FV.AMB_REWS,
                    FV.REWS,
                    FV.AMB_TI,
                    FV.TI,
                    FV.AMB_P,
                    FV.P,
                    FV.EFF,
                ]
            ]
        )
        print()

        # results by turbine
        turbine_results = o.reduce_states(
            {
                FV.AMB_P: "weights",
                FV.P: "weights",
                FV.AMB_CAP: "weights",
                FV.CAP: "weights",
            }
        )
        turbine_results[FV.AMB_YLD] = o.calc_turbine_yield(
            algo=algo, annual=True, ambient=True
        )
        turbine_results[FV.YLD] = o.calc_turbine_yield(algo=algo, annual=True)
        turbine_results[FV.EFF] = turbine_results[FV.P] / turbine_results[FV.AMB_P]
        print("\nResults by turbine:\n")
        print(turbine_results)

        # power results
        P0 = o.calc_mean_farm_power(ambient=True)
        P = o.calc_mean_farm_power()
        print(f"\nFarm power        : {P / 1000:.1f} MW")
        print(f"Farm ambient power: {P0 / 1000:.1f} MW")
        print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")
        print(f"Annual farm yield : {turbine_results[FV.YLD].sum():.2f} GWh.")
        print()

        if not args.nofig and args.calc_cline:
            points = np.zeros((n_s, n_p, 3))
            points[:, :, 0] = np.linspace(p0[0], p0[0] + n_t * stp[0] + 10 * D, n_p)[
                None, :
            ]
            points[:, :, 1] = p0[1]
            points[:, :, 2] = H
            print("\nPOINTS:\n", points[0])

            time0 = time.time()
            point_results = algo.calc_points(farm_results, points)
            time1 = time.time()
            print("\nCalc time =", time1 - time0, "\n")

            print(point_results)

            fig, ax = plt.subplots()
            for s in range(points.shape[0]):
                ax.plot(points[s, :, 0], point_results[FV.WS][s, :])
            ax.set_xlabel("x [m]")
            ax.set_ylabel("Wind speed [m/s]")
            ax.set_title("Centreline wind speed")
            plt.show()
            plt.close(fig)

        if not args.nofig and args.calc_mean:
            o = foxes.output.FlowPlots2D(algo, farm_results)
            fig = o.get_mean_fig_xy(FV.WS, resolution=10)
            plt.show()



import numpy as np

from foxes.core import Engine
import foxes.constants as FC


class DefaultEngine(Engine):
    """
    The case size dependent default engine.

    :group: engines

    """

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        self.finalize()

        with Engine.new(
            "process",
            n_procs=self.n_procs,
            chunk_size_states=self.chunk_size_states,
            chunk_size_points=self.chunk_size_points,
            verbosity=self.verbosity,
        ) as e:
            results = e.submit(f, *args, **kwargs)

        self.initialize()

        return results

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        self.finalize()

        with Engine.new(
            "process",
            n_procs=self.n_procs,
            chunk_size_states=self.chunk_size_states,
            chunk_size_points=self.chunk_size_points,
            verbosity=self.verbosity,
        ) as e:
            results = e.await_result(future)

        self.initialize()

        return results

    def map(
        self,
        func,
        inputs,
        *args,
        **kwargs,
    ):
        """
        Runs a function on a list of files

        Parameters
        ----------
        func: Callable
            Function to be called on each file,
            func(input, *args, **kwargs) -> data
        inputs: array-like
            The input data list
        args: tuple, optional
            Arguments for func
        kwargs: dict, optional
            Keyword arguments for func

        Returns
        -------
        results: list
            The list of results

        """

        self.finalize()

        with Engine.new(
            "process",
            n_procs=self.n_procs,
            chunk_size_states=self.chunk_size_states,
            chunk_size_points=self.chunk_size_points,
            verbosity=self.verbosity,
        ) as e:
            results = e.map(func, inputs, *args, **kwargs)

        self.initialize()

        return results

    def run_calculation(
        self,
        algo,
        model,
        model_data,
        farm_data=None,
        point_data=None,
        **kwargs,
    ):
        """
        Runs the model calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model: foxes.core.DataCalcModel
            The model that whose calculate function
            should be run
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset, optional
            The initial farm data
        point_data: xarray.Dataset, optional
            The initial point data

        Returns
        -------
        results: xarray.Dataset
            The model results

        """
        max_n = np.sqrt(self.n_workers) * (500 / algo.n_turbines) ** 1.5

        if (algo.n_states >= max_n) or (
            point_data is not None
            and algo.n_states * point_data.sizes[FC.TARGET] > 10000
        ):
            ename = "process"
        else:
            ename = "single"

        self.print(f"{type(self).__name__}: Selecting engine '{ename}'", level=1)

        self.finalize()

        with Engine.new(
            ename,
            n_procs=self.n_procs,
            chunk_size_states=self.chunk_size_states,
            chunk_size_points=self.chunk_size_points,
            verbosity=self.verbosity,
        ) as e:
            results = e.run_calculation(
                algo, model, model_data, farm_data, point_data=point_data, **kwargs
            )

        self.initialize()

        return results



from foxes.core import PartialWakesModel
import foxes.constants as FC


class RotorPoints(PartialWakesModel):
    """
    Partial wakes calculation directly by the
    rotor model.

    :group: models.partial_wakes

    """

    def get_wake_points(self, algo, mdata, fdata):
        """
        Get the wake calculation points, and their
        weights.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        rpoints: numpy.ndarray
            The wake calculation points, shape:
            (n_states, n_turbines, n_tpoints, 3)
        rweights: numpy.ndarray
            The target point weights, shape: (n_tpoints,)

        """
        return (
            algo.get_from_chunk_store(FC.ROTOR_POINTS, mdata=mdata),
            algo.get_from_chunk_store(FC.ROTOR_WEIGHTS, mdata=mdata),
        )

    def map_rotor_results(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        variable,
        rotor_res,
        rotor_weights,
    ):
        """
        Map ambient rotor point results onto target points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        variable: str
            The variable name to map
        rotor_res: numpy.ndarray
            The results at rotor points, shape:
            (n_states, n_turbines, n_rotor_points)
        rotor_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)

        Returns
        -------
        res: numpy.ndarray
            The mapped results at target points, shape:
            (n_states, n_targets, n_tpoints)

        """
        return rotor_res

    def finalize_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        rpoint_weights,
        wake_deltas,
        wmodel,
        downwind_index,
    ):
        """
        Updates the wake_deltas at the selected target
        downwind index.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.Data
            The target point data
        rpoint_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: np.ndarray of shape
            (n_states, n_turbines, n_tpoints)
        wmodel: foxes.core.WakeModel
            The wake model
        downwind_index: int
            The index in the downwind order

        Returns
        -------
        final_wake_deltas: dict
            The final wake deltas at the selected downwind
            turbines. Key: variable name, value: np.ndarray
            of shape (n_states, n_rotor_points)

        """
        wdel = {
            v: d[:, downwind_index, None].copy() if d.shape[1] > 1 else d[:, 0, None]
            for v, d in wake_deltas.items()
        }
        wmodel.finalize_wake_deltas(algo, mdata, fdata, tdata, wdel)

        return {v: d[:, 0] for v, d in wdel.items()}



from xarray import Dataset

from foxes.config import config
from foxes.utils import write_nc
import foxes.constants as FC

from .output import Output


class StateTurbineTable(Output):
    """
    Creates tables of state-turbine type data

    Attributes
    ----------
    farm_results: xarray.Dataset
        The farm results

    :group: output

    """

    def __init__(self, farm_results, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.farm_results = farm_results

    def get_dataset(
        self,
        variables,
        name_map={},
        to_file=None,
        isel=None,
        sel=None,
        **kwargs,
    ):
        """
        Creates a dataset object

        Parameters
        ----------
        variables: list of str
            The output variables
        name_map: dict
            Map from foxes to output names
        to_file: str, optional
            Name of the output file, if writing is desired
        isel: dict, optional
            Parameters for xarray.Dataset.isel
        sel: dict, optional
            Parameters for xarray.Dataset.sel
        kwargs: dict, optional
            Additional parameters for write_nc

        Returns
        -------
        table: xarray.Dataset
            The state-turbine data table

        """
        state = name_map.get(FC.STATE, FC.STATE)
        turbine = name_map.get(FC.TURBINE, FC.TURBINE)

        ds = Dataset(
            coords={
                state: self.farm_results[FC.STATE].to_numpy(),
                turbine: self.farm_results[FC.TURBINE].to_numpy(),
            },
            data_vars={
                name_map.get(v, v): ((state, turbine), self.farm_results[v].to_numpy())
                for v in variables
            },
        )

        if isel is not None:
            ds = ds.isel(**isel)
        if sel is not None:
            ds = ds.sel(**sel)

        if to_file is not None:
            fpath = self.get_fpath(to_file)
            write_nc(ds=ds, fpath=fpath, nc_engine=config.nc_engine, **kwargs)

        return ds



import numpy as np

from foxes.core import WindVectorWakeSuperposition
from foxes.utils import wd2uv, uv2wd, delta_wd
import foxes.variables as FV
import foxes.constants as FC


class WindVectorLinear(WindVectorWakeSuperposition):
    """
    Linear superposition of wind deficit vector results

    Attributes
    ----------
    scale_amb: bool
        Flag for scaling wind deficit with ambient wind speed
        instead of waked wind speed

    :group: models.wake_superpositions

    """

    def __init__(self, scale_amb=False):
        """
        Constructor.

        Parameters
        ----------
        scale_amb: bool
            Flag for scaling wind deficit with ambient wind speed
            instead of waked wind speed

        """
        super().__init__()
        self.scale_amb = scale_amb

    def __repr__(self):
        a = f"scale_amb={self.scale_amb}"
        return f"{type(self).__name__}({a})"

    def input_farm_vars(self, algo):
        """
        The variables which are needed for running
        the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        input_vars: list of str
            The input variable names

        """
        return [FV.AMB_REWS] if self.scale_amb else [FV.REWS]

    def wdeltas_ws2uv(self, algo, fdata, tdata, downwind_index, wdeltas, st_sel):
        """
        Transform results from wind speed to wind vector data

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_tpoints)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, now respecting has_uv flag

        """
        if FV.AMB_UV not in tdata:
            tdata[FV.AMB_UV] = wd2uv(tdata[FV.AMB_WD], tdata[FV.AMB_WS])
        if FV.UV not in wdeltas:
            assert FV.WS in wdeltas, (
                f"{self.name}: Expecting '{FV.WS}' in wdeltas, got {list(wdeltas.keys())}"
            )
            scale = self.get_data(
                FV.AMB_REWS if self.scale_amb else FV.REWS,
                FC.STATE_TARGET_TPOINT,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=False,
                selection=st_sel,
            )
            ws0 = tdata[FV.AMB_WS][st_sel]
            wd0 = tdata[FV.AMB_WD][st_sel]
            dws = scale * wdeltas.pop(FV.WS)
            dwd = wdeltas.pop(FV.WD, 0)
            wdeltas[FV.UV] = wd2uv(wd0 + dwd, ws0 + dws) - tdata[FV.AMB_UV][st_sel]

        return wdeltas

    def wdeltas_uv2ws(self, algo, fdata, tdata, downwind_index, wdeltas, st_sel):
        """
        Transform results from wind vector to wind speed data

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_tpoints)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, now respecting has_uv flag

        """
        if FV.UV in wdeltas:
            scale = self.get_data(
                FV.AMB_REWS if self.scale_amb else FV.REWS,
                FC.STATE_TARGET_TPOINT,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=False,
                selection=st_sel,
            )
            ws0 = tdata[FV.AMB_WS][st_sel]
            wd0 = tdata[FV.AMB_WD][st_sel]
            uv = tdata[FV.AMB_UV][st_sel] + wdeltas.pop(FV.UV)
            wdeltas[FV.WD] = delta_wd(wd0, uv2wd(uv))
            wdeltas[FV.WS] = (np.linalg.norm(uv, axis=-1) - ws0) / scale

        return wdeltas

    def add_wake_vector(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        wake_delta_uv,
        wake_model_result_uv,
    ):
        """
        Add a wake delta vector to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        wake_delta_uv: numpy.ndarray
            The original wind vector wake deltas, shape:
            (n_states, n_targets, n_tpoints, 2)
        wake_model_result_uv: numpy.ndarray
            The new wind vector wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, 2, ...)

        Returns
        -------
        wdelta_uv: numpy.ndarray
            The updated wind vector wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """

        if np.any(st_sel):
            wake_delta_uv[st_sel] += wake_model_result_uv

        return wake_delta_uv

    def calc_final_wake_delta_uv(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        wake_delta_uv,
    ):
        """
        Calculate the final wind vector wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        wake_delta_uv: numpy.ndarray
            The original wind vector wake deltas, shape:
            (n_states, n_targets, n_tpoints, 2)

        Returns
        -------
        final_wake_delta_ws: numpy.ndarray
            The final wind speed wake delta, which will be added to
            the ambient results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)
        final_wake_delta_wd: numpy.ndarray
            The final wind direction wake delta, which will be added to
            the ambient results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        if FV.AMB_UV not in tdata:
            tdata[FV.AMB_UV] = wd2uv(tdata[FV.AMB_WD], tdata[FV.AMB_WS])

        uv = tdata[FV.AMB_UV] + wake_delta_uv
        dwd = delta_wd(tdata[FV.AMB_WD], uv2wd(uv))
        dws = np.linalg.norm(uv, axis=-1) - tdata[FV.AMB_WS]

        return dws, dwd



import numpy as np

from foxes.config import config
from foxes.utils import get_utm_zone, from_lonlat


class WindFarm:
    """
    The wind farm.

    Attributes
    ----------
    name: str
        The wind farm name
    turbines: list of foxes.core.Turbine
        The wind turbines
    boundary: foxes.utils.geom2d.AreaGeometry, optional
        The wind farm boundary

    :group: core

    """

    def __init__(
        self,
        name="wind_farm",
        boundary=None,
        input_is_lonlat=False,
        utm_zone="from_farm",
    ):
        """
        Constructor.

        Parameters
        ----------
        name: str
            The wind farm name
        boundary: foxes.utils.geom2d.AreaGeometry, optional
            The wind farm boundary
        input_is_lonlat: bool, optional
            Whether the input coordinates are given in lon, lat. If True,
            the coordinates are converted to UTM as specified by the
            utm_zone parameter.
        utm_zone: str or tuple, optional
            Method for setting UTM zone in config, if not already set.
            Options are:
            - "from_turbine_X": use turbine X coordinates
            - "from_farm": use farm center coordinates
            - "XA": use given number X, letter A
            - (lon, lat): use given lon, lat values
            - None: do not set UTM zone, assume it is already set

        """
        self.name = name
        self.__turbines = []
        self.boundary = boundary

        self.__data_is_lonlat = input_is_lonlat
        self.__utm_zone = utm_zone
        self.__locked = False

    @property
    def data_is_lonlat(self):
        """
        Whether the input coordinates are given in lat, lon.

        Returns
        -------
        data_is_lonlat: bool
            True if the input coordinates are given in lat, lon

        """
        return self.__data_is_lonlat

    @property
    def locked(self):
        """
        Whether the wind farm is locked (no more turbines can be added)

        Returns
        -------
        locked: bool
            True if the wind farm is locked

        """
        return self.__locked

    @property
    def turbines(self):
        """
        The list of wind turbines

        Returns
        -------
        turbines: list of foxes.core.Turbine
            The wind turbines

        """
        if not self.__locked:
            self.__locked = True
            if self.__data_is_lonlat:
                if not config.utm_zone_set and self.__utm_zone is None:
                    raise ValueError(
                        f"WindFarm '{self.name}': input_is_lonlat is True, but config.utm_zone and utm_zone are None"
                    )
                if self.__utm_zone is None:
                    zone = config.utm_zone
                elif self.__utm_zone == "from_farm":
                    lonlat = np.mean([t.xy for t in self.__turbines], axis=0)
                    zone = get_utm_zone(lonlat[None, :])
                elif (
                    isinstance(self.__utm_zone, str)
                    and self.__utm_zone.startswith("from_turbine_")
                    and len(self.__utm_zone) > len("from_turbine_")
                ):
                    idx = int(self.__utm_zone[len("from_turbine_") :])
                    lonlat = self.__turbines[idx].xy
                    zone = get_utm_zone(lonlat[None, :])
                elif isinstance(self.__utm_zone, str):
                    zone = (int(self.__utm_zone[:-1]), self.__utm_zone[-1])
                elif len(self.__utm_zone) == 2:
                    lonlat = np.asarray(self.__utm_zone)
                    zone = get_utm_zone(lonlat[None, :])
                else:
                    raise ValueError(
                        f"WindFarm '{self.name}': invalid utm_zone argument: {self.__utm_zone}"
                    )
                if not config.utm_zone_set:
                    config.set_utm_zone(*zone)
                elif config.utm_zone != zone:
                    raise ValueError(
                        f"WindFarm '{self.name}': input_is_lonlat is True, but config.utm_zone = {config.utm_zone} differs from determined zone {zone}"
                    )
                for t in self.__turbines:
                    t.xy = from_lonlat(t.xy[None, :])[0]
                self.__data_is_lonlat = False
        return self.__turbines

    def lock(self, verbosity=1):
        """
        Lock the wind farm (no more turbines can be added)

        Parameters
        ----------
        verbosity: int
            The output verbosity, 0 = silent

        """
        self.turbines
        if verbosity > 0:
            if config.utm_zone_set:
                utmn, utml = config.utm_zone
                print(
                    f"WindFarm '{self.name}': locked with {self.n_turbines} turbines, UTM zone {utmn}{utml}"
                )
                if verbosity > 1:
                    for t in self.__turbines:
                        print(
                            f"  Turbine {t.index}, {t.name}: UTM {utmn}{utml}, xy=({t.xy[0]:.2f}, {t.xy[1]:.2f}), {', '.join(t.models)}"
                        )
            else:
                print(f"WindFarm '{self.name}': locked with {self.n_turbines} turbines")

    def add_turbine(self, turbine, verbosity=1):
        """
        Add a wind turbine to the list.

        Parameters
        ----------
        turbine: foxes.core.Turbine
            The wind turbine
        verbosity: int
            The output verbosity, 0 = silent

        """
        assert not self.__locked, (
            f"WindFarm '{self.name}': cannot add turbine, farm is locked"
        )
        if turbine.index is None:
            turbine.index = len(self.__turbines)
        if turbine.name is None:
            turbine.name = f"T{turbine.index}"
        self.__turbines.append(turbine)
        if verbosity > 0:
            if self.data_is_lonlat:
                print(
                    f"Turbine {turbine.index}, {turbine.name}: lonlat=({turbine.xy[0]:.6f}, {turbine.xy[1]:.6f}), {', '.join(turbine.models)}"
                )
            elif config.utm_zone_set:
                utmn, utml = config.utm_zone
                print(
                    f"Turbine {turbine.index}, {turbine.name}: UTM {utmn}{utml}, xy=({turbine.xy[0]:.2f}, {turbine.xy[1]:.2f}), {', '.join(turbine.models)}"
                )
            else:
                print(
                    f"Turbine {turbine.index}, {turbine.name}: xy=({turbine.xy[0]:.2f}, {turbine.xy[1]:.2f}), {', '.join(turbine.models)}"
                )

    @property
    def n_turbines(self):
        """
        The number of turbines in the wind farm

        Returns
        -------
        n_turbines: int
            The total number of turbines

        """
        return len(self.__turbines)

    @property
    def turbine_names(self):
        """
        The list of names of all turbines

        Returns
        -------
        names: list of str
            The names of all turbines

        """
        return [t.name for t in self.__turbines]

    @property
    def xy_array(self):
        """
        Returns an array of the wind farm ground points

        Returns
        -------
        xya: numpy.ndarray
            The turbine ground positions, shape: (n_turbines, 2)

        """
        return np.array([t.xy for t in self.__turbines], dtype=config.dtype_double)

    @property
    def wind_farm_names(self):
        """
        The list of wind farm names for all turbines

        Returns
        -------
        names: list of str
            The wind farm names for all turbines

        """
        return list(
            set(
                [
                    t.wind_farm_name if t.wind_farm_name is not None else self.name
                    for t in self.__turbines
                ]
            )
        )

    def get_wind_farm_mapping(self):
        """
        Returns a mapping from wind farm names to turbine indices

        Returns
        -------
        mapping: dict
            A dictionary, where keys are wind farm names and
            values are lists of turbine indices belonging to that wind farm

        """
        mapping = {}
        for i, t in enumerate(self.__turbines):
            wf_name = t.wind_farm_name if t.wind_farm_name is not None else self.name
            if wf_name not in mapping:
                mapping[wf_name] = []
            mapping[wf_name].append(i)
        return mapping

    def get_xy_bounds(self, extra_space=None, algo=None):
        """
        Returns min max points of the wind farm ground points

        Parameters
        ----------
        extra_space: float or str, optional
            The extra space, either float in m,
            or str for units of D, e.g. '2.5D'
        algo: foxes.core.Algorithm, optional
            The algorithm

        Returns
        -------
        x_mima: numpy.ndarray
            The (x_min, x_max) point
        y_mima: numpy.ndarray
            The (y_min, y_max) point

        """
        if self.boundary is not None:
            xy = None
            p_min, p_max = self.boundary.p_min(), self.boundary.p_max()
        else:
            xy = self.xy_array
            p_min, p_max = np.min(xy, axis=0), np.max(xy, axis=0)

        if extra_space is not None:
            if isinstance(extra_space, str):
                assert algo is not None, (
                    f"WindFarm: require algo argument for extra_space '{extra_space}'"
                )
                assert len(extra_space) > 1 and extra_space[-1] == "D", (
                    f"Expecting float or str like '2.5D', got extra_space = '{extra_space}'"
                )
                extra_space = float(extra_space[:-1])
                rds = self.get_rotor_diameters(algo)
                if xy is None:
                    extra_space *= np.max(rds)
                else:
                    p_min = np.min(xy - extra_space * rds[:, None], axis=0)
                    p_max = np.max(xy + extra_space * rds[:, None], axis=0)
                    return p_min, p_max

            p_min -= extra_space
            p_max += extra_space

        return p_min, p_max

    def get_rotor_diameters(self, algo):
        """
        Gets the rotor diameters

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm

        Returns
        -------
        rds: numpy.ndarray
            The rotor diameters, shape: (n_turbienes,)

        """
        rds = [
            t.D if t.D is not None else algo.farm_controller.turbine_types[i].D
            for i, t in enumerate(self.__turbines)
        ]
        return np.array(rds, dtype=config.dtype_double)

    def get_hub_heights(self, algo):
        """
        Gets the hub heights

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm

        Returns
        -------
        hhs: numpy.ndarray
            The hub heights, shape: (n_turbienes,)

        """
        hhs = [
            t.H if t.H is not None else algo.farm_controller.turbine_types[i].H
            for i, t in enumerate(self.__turbines)
        ]
        return np.array(hhs, dtype=config.dtype_double)

    def get_capacity(self, algo):
        """
        Gets the total capacity of the wind farm

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm

        Returns
        -------
        capa: float
            The total capacity in W

        """
        ttypes = algo.farm_controller.turbine_types
        assert ttypes is not None, (
            f"WindFarm '{self.name}': turbine types not set in farm controller {algo.farm_controller.name}"
        )

        cap = 0.0
        for tt in ttypes:
            assert tt.P_nominal is not None, (
                f"WindFarm '{self.name}': P_nominal not set for turbine type '{tt.name}' "
            )
            cap += tt.P_nominal
        return cap



"""
Turbine type models.
"""

from .PCt_file import PCtFile as PCtFile
from .PCt_from_two import PCtFromTwo as PCtFromTwo
from .CpCt_file import CpCtFile as CpCtFile
from .CpCt_from_two import CpCtFromTwo as CpCtFromTwo
from .null_type import NullType as NullType
from .wsrho2PCt_from_two import WsRho2PCtFromTwo as WsRho2PCtFromTwo
from .wsti2PCt_from_two import WsTI2PCtFromTwo as WsTI2PCtFromTwo
from .lookup import FromLookupTable as FromLookupTable
from .TBL_file import TBLFile as TBLFile
from .calculator_type import CalculatorType as CalculatorType



"""
General utilities.
"""

from .wind_dir import wd2uv as wd2uv
from .wind_dir import wd2wdvec as wd2wdvec
from .wind_dir import wdvec2wd as wdvec2wd
from .wind_dir import uv2wd as uv2wd
from .wind_dir import delta_wd as delta_wd

from .subclasses import all_subclasses as all_subclasses
from .subclasses import new_cls as new_cls
from .subclasses import new_instance as new_instance

from .factory import Factory as Factory
from .factory import FDict as FDict
from .factory import WakeKFactory as WakeKFactory

from .geopandas_utils import read_shp as read_shp
from .geopandas_utils import shp2csv as shp2csv
from .geopandas_utils import read_shp_polygons as read_shp_polygons
from .geopandas_utils import shp2geom2d as shp2geom2d

from .utm_utils import to_lonlat as to_lonlat
from .utm_utils import from_lonlat as from_lonlat
from .utm_utils import get_utm_zone as get_utm_zone

from .load import import_module as import_module
from .load import load_module as load_module

from .xarray_utils import write_nc as write_nc
from .xarray_utils import compute_scale_and_offset as compute_scale_and_offset
from .xarray_utils import pack_value as pack_value
from .xarray_utils import unpack_value as unpack_value
from .xarray_utils import get_encoding as get_encoding

from .pandas_utils import PandasFileHelper as PandasFileHelper
from .dict import Dict as Dict
from .data_book import DataBook as DataBook
from .cubic_roots import cubic_roots as cubic_roots
from .exec_python import exec_python as exec_python
from .regularize import sqrt_reg as sqrt_reg
from .tab_files import read_tab_file as read_tab_file
from .random_xy import random_xy_square as random_xy_square
from .dev_utils import print_mem as print_mem
from .wrg_utils import ReaderWRG as ReaderWRG
from .weibull import weibull_weights as weibull_weights

from . import two_circles as two_circles
from . import abl as abl
from . import geom2d as geom2d



"""
Partial wake models.
"""

from .centre import PartialCentre as PartialCentre
from .rotor_points import RotorPoints as RotorPoints
from .top_hat import PartialTopHat as PartialTopHat
from .axiwake import PartialAxiwake as PartialAxiwake
from .segregated import PartialSegregated as PartialSegregated
from .grid import PartialGrid as PartialGrid



from foxes.utils import import_module

from .pool import PoolEngine


class MultiprocessEngine(PoolEngine):
    """
    The multiprocessing engine for foxes calculations.

    :group: engines

    """

    def _create_pool(self):
        """Creates the pool"""
        Pool = import_module("multiprocess").Pool
        self._pool = Pool(processes=self.n_workers)

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        return self._pool.apply_async(f, args=args, kwds=kwargs)

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        return future.get()

    def _shutdown_pool(self):
        """Shuts down the pool"""
        self._pool.close()
        self._pool.terminate()
        self._pool.join()



from .random_abl_states import create_random_abl_states as create_random_abl_states
from .random_abl_states import write_random_abl_states as write_random_abl_states
from .random_timeseries import random_timseries_data as random_timseries_data



import xarray as xr
import pandas as pd

import foxes.constants as FC

from .output import Output
from .slice_data import SliceData


class SlicesData(Output):
    """
    Create data for horizontal or vertical 2D slices, all in a
    single Dataset

    :group: output

    """

    def __init__(
        self,
        algo,
        farm_results,
        verbosity_delta=1,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        algo: foxes.Algorithm
            The algorithm for point calculation
        farm_results: xarray.Dataset
            The farm results
        verbosity_delta: int
            Verbosity threshold for printing calculation info
        kwargs: dict, optional
            Additional parameters for the Output class

        """
        super().__init__(**kwargs)
        self._slice_data = SliceData(
            algo=algo,
            farm_results=farm_results,
            verbosity_delta=verbosity_delta,
            **kwargs,
        )

    def get_mean_data_xy(
        self,
        z_list,
        *args,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates mean data of 2D farm flow slices in a xy-plane.

        Parameters
        ----------
        z_list: list of float
            The z values
        args: tuple, optional
            Arguments for the SliceData function of the same name
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Arguments for the SliceData function of the same name

        Returns
        -------
        data: xarray.Dataset
            The gridded data

        """
        dsl = []
        for z in z_list:
            if verbosity > 0:
                print(f"{type(self).__name__}: Creating slice z = {z}")
            dsl.append(
                self._slice_data.get_mean_data_xy(
                    *args,
                    z=z,
                    data_format="xarray",
                    ret_states=False,
                    ret_grid=False,
                    verbosity=verbosity,
                    **kwargs,
                )
            )
        out = xr.concat(dsl, pd.Index(z_list, name="z"))
        del out.attrs["z"]
        return out

    def get_mean_data_xz(
        self,
        y_list,
        *args,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates mean data of 2D farm flow slices in a xz-plane.

        Parameters
        ----------
        y_list: list of float
            The y values
        args: tuple, optional
            Arguments for the SliceData function of the same name
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Arguments for the SliceData function of the same name

        Returns
        -------
        data: xarray.Dataset
            The gridded data

        """
        dsl = []
        for y in y_list:
            if verbosity > 0:
                print(f"{type(self).__name__}: Creating slice y = {y}")
            dsl.append(
                self._slice_data.get_mean_data_xz(
                    *args,
                    y=y,
                    data_format="xarray",
                    ret_states=False,
                    ret_grid=False,
                    verbosity=verbosity,
                    **kwargs,
                )
            )
        out = xr.concat(dsl, pd.Index(y_list, name="y"))
        del out.attrs["y"]
        return out

    def get_mean_data_yz(
        self,
        x_list,
        *args,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates mean data of 2D farm flow slices in a yz-plane.

        Parameters
        ----------
        x_list: list of float
            The x values
        args: tuple, optional
            Arguments for the SliceData function of the same name
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Arguments for the SliceData function of the same name

        Returns
        -------
        data: xarray.Dataset
            The gridded data

        """
        dsl = []
        for x in x_list:
            if verbosity > 0:
                print(f"{type(self).__name__}: Creating slice x = {x}")
            dsl.append(
                self._slice_data.get_mean_data_yz(
                    *args,
                    x=x,
                    data_format="xarray",
                    ret_states=False,
                    ret_grid=False,
                    verbosity=verbosity,
                    **kwargs,
                )
            )
        out = xr.concat(dsl, pd.Index(x_list, name="x"))
        del out.attrs["x"]
        return out

    def get_states_data_xy(
        self,
        z_list,
        *args,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates states data of 2D farm flow slices in a xy-plane.

        Parameters
        ----------
        z_list: list of float
            The z values
        args: tuple, optional
            Arguments for the SliceData function of the same name
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Arguments for the SliceData function of the same name

        Returns
        -------
        data: xarray.Dataset
            The gridded data

        """
        dsl = []
        for z in z_list:
            if verbosity > 0:
                print(f"{type(self).__name__}: Creating slice z = {z}")
            dsl.append(
                self._slice_data.get_states_data_xy(
                    *args,
                    z=z,
                    data_format="xarray",
                    ret_states=False,
                    ret_grid=False,
                    verbosity=verbosity,
                    **kwargs,
                )
            )
        out = xr.concat(dsl, pd.Index(z_list, name="z"))
        del out.attrs["z"]
        return out.transpose(FC.STATE, "z", ...)

    def get_states_data_xz(
        self,
        y_list,
        *args,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates states data of 2D farm flow slices in a xz-plane.

        Parameters
        ----------
        y_list: list of float
            The y values
        args: tuple, optional
            Arguments for the SliceData function of the same name
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Arguments for the SliceData function of the same name

        Returns
        -------
        data: xarray.Dataset
            The gridded data

        """
        dsl = []
        for y in y_list:
            if verbosity > 0:
                print(f"{type(self).__name__}: Creating slice y = {y}")
            dsl.append(
                self._slice_data.get_states_data_xz(
                    *args,
                    y=y,
                    data_format="xarray",
                    ret_states=False,
                    ret_grid=False,
                    verbosity=verbosity,
                    **kwargs,
                )
            )
        out = xr.concat(dsl, pd.Index(y_list, name="y"))
        del out.attrs["y"]
        return out.transpose(FC.STATE, "z", "y", ...)

    def get_states_data_yz(
        self,
        x_list,
        *args,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates states data of 2D farm flow slices in a yz-plane.

        Parameters
        ----------
        x_list: list of float
            The x values
        args: tuple, optional
            Arguments for the SliceData function of the same name
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Arguments for the SliceData function of the same name

        Returns
        -------
        data: xarray.Dataset
            The gridded data

        """
        dsl = []
        for x in x_list:
            if verbosity > 0:
                print(f"{type(self).__name__}: Creating slice x = {x}")
            dsl.append(
                self._slice_data.get_states_data_yz(
                    *args,
                    x=x,
                    data_format="xarray",
                    ret_states=False,
                    ret_grid=False,
                    verbosity=verbosity,
                    **kwargs,
                )
            )
        out = xr.concat(dsl, pd.Index(x_list, name="x"))
        del out.attrs["x"]
        return out.transpose(FC.STATE, "z", "y", "x", ...)



"""
Turbine models.
"""

from .kTI_model import kTI as kTI
from .set_farm_vars import SetFarmVars as SetFarmVars
from .thrust2ct import Thrust2Ct as Thrust2Ct
from .power_mask import PowerMask as PowerMask
from .sector_management import SectorManagement as SectorManagement
from .table_factors import TableFactors as TableFactors
from .yaw2yawm import YAW2YAWM as YAW2YAWM
from .yawm2yaw import YAWM2YAW as YAWM2YAW
from .calculator import Calculator as Calculator
from .rotor_centre_calc import RotorCentreCalc as RotorCentreCalc
from .lookup_table import LookupTable as LookupTable



import numpy as np
import utm


def to_lonlat(xy):
    """
    Convert UTM coordinates to longitude and latitude.

    Parameters
    ----------
    xy: array_like
        The cartesian coordinates, shape: (n_points, 2), where
        the last dimension is (x, y).

    Returns
    -------
    lonlat: numpy.ndarray
        The longitude and latitude coordinates, shape: (n_points, 2), where
        the last dimension is (lon, lat).

    :group: utils

    """
    from foxes.config.conf import config

    utmz, utml = config.utm_zone
    lat, lon = utm.to_latlon(
        xy[:, 0],
        xy[:, 1],
        utmz,
        utml,
    )
    return np.stack((lon, lat), axis=-1)


def from_lonlat(lonlat):
    """
    Convert longitude and latitude to UTM coordinates.

    Parameters
    ----------
    lonlat: array_like
        The longitude and latitude coordinates, shape: (n_points, 2), where
        the last dimension is (lon, lat).

    Returns
    -------
    xy: numpy.ndarray
        The cartesian coordinates, shape: (n_points, 2), where
        the last dimension is (x, y).

    :group: utils

    """
    from foxes.config import config

    utmz, utml = config.utm_zone
    x, y, __, __ = utm.from_latlon(
        lonlat[:, 1].astype(np.float64),
        lonlat[:, 0].astype(np.float64),
        utmz,
        utml,
    )
    return np.stack((x, y), axis=-1)


def get_utm_zone(lonlat):
    """
    Get the UTM zone for given latitude and longitude.

    Parameters
    ----------
    lonlat: array_like
        The latitude and longitude coordinates, shape: (n_points, 2), where
        the last dimension is (lat, lon).

    Returns
    -------
    utm_zone: tuple
        The UTM zone as (zone_number, zone_letter).

    :group: utils

    """
    lat = lonlat[:, 1].astype(np.float64)
    lon = lonlat[:, 0].astype(np.float64)
    return utm.from_latlon(lat, lon)[2:4]



import time
import argparse
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path

import foxes
import foxes.variables as FV


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a", "--animation", help="Write flow animation file", action="store_true"
    )
    parser.add_argument(
        "-A",
        "--ani_file",
        help="Path to the animation file to be written",
        default="ani.gif",
    )
    parser.add_argument(
        "-F",
        "--fps",
        help="The frames per second value for the animation",
        type=int,
        default=4,
    )
    parser.add_argument(
        "-b",
        "--background",
        help="Switch on dynamic background interpretation",
        action="store_true",
    )
    parser.add_argument(
        "-S",
        "--max_state",
        help="States subset to the first n states",
        type=int,
        default=None,
    )
    parser.add_argument(
        "-R",
        "--ref_xy",
        help="The reference point x y",
        type=float,
        default=[2500, 2500],
        nargs="+",
    )
    parser.add_argument(
        "-nt", "--n_turbines", help="The number of turbines", default=9, type=int
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The timeseries input file (path or static)",
        default="timeseries_100.csv.gz",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear_lim_k004"],
        nargs="+",
    )
    parser.add_argument("-f", "--frame", help="The wake frame", default="timelines")
    parser.add_argument(
        "-d", "--deflection", help="The wake deflection", default="no_deflection"
    )
    parser.add_argument(
        "-y",
        "--yawm",
        help="The uniform yaw misalignment value",
        type=float,
        default=None,
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-sl",
        "--show_layout",
        help="Flag for showing layout figure",
        action="store_true",
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=20,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=5000,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    # optionally set turbines in yaw:
    N = int(args.n_turbines**0.5)
    if args.yawm is None:
        ymodels = []
    else:
        yawm = np.zeros((1, N * N), dtype=np.float64)
        yawm[:, :N] = args.yawm
        mbook.turbine_models["set_yawm"] = foxes.models.turbine_models.SetFarmVars(
            pre_rotor=True
        )
        mbook.turbine_models["set_yawm"].add_var(FV.YAWM, yawm)
        ymodels = ["set_yawm"]

    if not args.background:
        States = foxes.input.states.Timeseries
        kwargs = {}
    else:
        States = foxes.input.states.OnePointFlowTimeseries
        kwargs = {"ref_xy": args.ref_xy}

    states = States(
        data_source=args.states,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti"},
        fixed_vars={FV.RHO: 1.225, FV.TI: 0.07},
        states_sel=range(args.max_state) if args.max_state is not None else None,
        **kwargs,
    )

    farm = foxes.WindFarm()
    N = int(args.n_turbines**0.5)
    foxes.input.farm_layout.add_grid(
        farm,
        xy_base=np.array([0.0, 0.0]),
        step_vectors=np.array([[1000.0, 0], [0, 800.0]]),
        steps=(N, N),
        turbine_models=ymodels + args.tmodels + [ttype.name],
    )

    if not args.nofig and args.show_layout:
        ax = foxes.output.FarmLayoutOutput(farm).get_figure()
        plt.show()
        plt.close(ax.get_figure())

    algo = foxes.algorithms.Iterative(
        farm,
        states=states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame=args.frame,
        wake_deflection=args.deflection,
        partial_wakes=args.pwakes,
        mbook=mbook,
        verbosity=1,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()

        print("\nCalc time =", time1 - time0, "\n")

        o = foxes.output.FarmResultsEval(farm_results)
        o.add_capacity(algo)
        o.add_capacity(algo, ambient=True)
        o.add_efficiency()

        print("\nFarm results:\n")
        print(farm_results)

        # state-turbine results
        farm_df = farm_results.to_dataframe()
        print("\nFarm results data:\n")
        print(
            farm_df[
                [
                    FV.X,
                    FV.Y,
                    FV.WD,
                    FV.AMB_REWS,
                    FV.REWS,
                    FV.AMB_TI,
                    FV.TI,
                    FV.AMB_P,
                    FV.P,
                    FV.EFF,
                ]
            ]
        )
        print()
        print(farm_df[[FV.AMB_REWS, FV.REWS, FV.CT, FV.EFF]].describe())

        # power results
        P0 = o.calc_mean_farm_power(ambient=True)
        P = o.calc_mean_farm_power()
        print(f"\nFarm power        : {P / 1000:.1f} MW")
        print(f"Farm ambient power: {P0 / 1000:.1f} MW")
        print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")

        if not args.nofig:
            sts = np.arange(farm_results.sizes["state"])
            plt.plot(sts, farm_results.REWS[:, 4], label="Turbine 4")
            plt.plot(sts, farm_results.REWS[:, 7], label="Turbine 7")
            plt.legend()
            plt.xlabel("State")
            plt.ylabel("REWS [m/s]")
            plt.show()

        if not args.nofig and args.animation:
            print("\nCalculating animation")

            fig, axs = plt.subplots(
                2, 1, figsize=(5.2, 7), gridspec_kw={"height_ratios": [3, 1]}
            )

            anim = foxes.output.Animator(fig)
            of = foxes.output.FlowPlots2D(algo, farm_results)
            anim.add_generator(
                of.gen_states_fig_xy(
                    FV.WS,
                    resolution=30,
                    quiver_pars=dict(angles="xy", scale_units="xy", scale=0.013),
                    quiver_n=35,
                    xmax=5000,
                    ymax=5000,
                    vmin=0,
                    fig=fig,
                    ax=axs[0],
                    ret_im=True,
                    title=None,
                    animated=True,
                    rotor_color="red",
                )
            )
            anim.add_generator(
                o.gen_stdata(
                    turbines=[4, 7],
                    variable=FV.REWS,
                    fig=fig,
                    ax=axs[1],
                    ret_im=True,
                    legloc="upper left",
                    animated=True,
                )
            )

            ani = anim.animate()

            lo = foxes.output.FarmLayoutOutput(farm)
            lo.get_figure(
                fig=fig,
                ax=axs[0],
                title="",
                annotate=1,
                anno_delx=-120,
                anno_dely=-60,
                alpha=0,
            )

            fpath = Path(args.ani_file)
            print("Writing file", fpath)
            if fpath.suffix == ".gif":
                ani.save(filename=fpath, writer="pillow", fps=args.fps)
            else:
                ani.save(filename=fpath, writer="ffmpeg", fps=args.fps)



from .conf import Config as Config
from .conf import config as config
from .conf import get_path as get_path
from .conf import get_input_path as get_input_path
from .conf import get_output_path as get_output_path



"""
Algorithms define the main calculation routines.
"""

from .downwind.downwind import Downwind as Downwind
from .iterative.iterative import Iterative as Iterative
from .sequential import Sequential as Sequential

from . import downwind as downwind
from . import iterative as iterative
from . import sequential as sequential



import numpy as np
import pandas as pd
from cycler import cycler
import matplotlib.pyplot as plt

from .output import Output
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class FarmResultsEval(Output):
    """
    Evaluates farm results data.

    This sums over turbines and/or states,
    given the state-turbine farm_calc results.

    Attributes
    ----------
    results: xarray.Dataset
        The farm results

    :group: output

    """

    def __init__(self, farm_results, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.results = farm_results

    def weinsum(self, rhs, *vars):
        """
        Calculates Einstein sum, adding weights
        as last argument to the given fields.

        It's all about treating NaN values.

        Parameters
        ----------
        rhs: str
            The right-hand side of the einsum expression.
            Convention: 's' for states, 't' for turbines
        vars: tuple of str or np.ndarray
            The variables mentioned in the expression,
            but without the obligatory weights that will
            be added at the end

        Returns
        -------
        result: np.ndarray
            The results array

        """
        nas = None
        fields = []
        for v in vars:
            if isinstance(v, str):
                vdata = self.results[v].to_numpy()
                nns = np.sum(np.isnan(vdata))
                assert nns == 0, (
                    f"Found {nns} nan values for variable '{v}' of shape {vdata.shape}"
                )
                fields.append(vdata)
            else:
                fields.append(v)
            if nas is None:
                nas = np.zeros_like(fields[-1], dtype=bool)
            nas = nas | np.isnan(fields[-1])

        inds = ["st" for __ in fields]
        if self.results[FV.WEIGHT].dims == (FC.STATE,):
            inds += ["s"]

            if np.any(nas):
                sel = ~np.any(nas, axis=1)
                fields = [f[sel] for f in fields]

                weights0 = self.results[FV.WEIGHT].to_numpy()
                w0 = np.sum(weights0)
                weights = weights0[sel]
                w1 = np.sum(weights)
                weights *= w0 / w1
                fields.append(weights)

            else:
                fields.append(self.results[FV.WEIGHT].to_numpy())

        elif self.results[FV.WEIGHT].dims == (FC.STATE, FC.TURBINE):
            inds += ["st"]

            if np.any(nas):
                sel = ~np.any(nas, axis=1)
                fields = [f[sel] for f in fields]

                weights0 = self.results[FV.WEIGHT].to_numpy()
                w0 = np.sum(weights0, axis=0)[None, :]
                weights = weights0[sel]
                w1 = np.sum(weights, axis=0)[None, :]
                weights *= w0 / w1
                fields.append(weights)

            else:
                fields.append(self.results[FV.WEIGHT].to_numpy())

        else:
            raise ValueError(
                f"Expecting '{FV.WEIGHT}' variable with dimensions {(FC.STATE,)} or {(FC.STATE, FC.TURBINE)}, got {self.results[FV.WEIGHT].dims}"
            )
        expr = ",".join(inds) + "->" + rhs

        return np.einsum(expr, *fields)

    def reduce_states(self, vars_op):
        """
        Reduces the states dimension by some operation

        Parameters
        ----------
        vars_op: dict
            The operation per variable. Key: str, the variable
            name. Value: str, the operation, choices
            are: weights, mean_no_weights, sum, min, max.

        Returns
        -------
        data: pandas.DataFrame
            The results per turbine

        """
        n_turbines = self.results.sizes[FC.TURBINE]

        rdata = {}
        for v, op in vars_op.items():
            vdata = self.results[v].to_numpy()
            nns = np.sum(np.isnan(vdata))
            assert nns == 0, (
                f"Found {nns} nan values for variable '{v}' of shape {vdata.shape}"
            )

            if op == "weights":
                rdata[v] = self.weinsum("t", vdata)
            elif op == "mean_no_weights":
                rdata[v] = np.mean(vdata, axis=0)
            elif op == "sum":
                rdata[v] = np.sum(vdata, axis=0)
            elif op == "min":
                rdata[v] = np.min(vdata, axis=0)
            elif op == "max":
                rdata[v] = np.max(vdata, axis=0)
            elif op == "std":
                rdata[v] = np.std(vdata, axis=0)
            else:
                raise KeyError(
                    f"Unknown operation '{op}' for variable '{v}'. Please choose: weights, mean_no_weights, sum, min, max"
                )

        data = pd.DataFrame(index=range(n_turbines), data=rdata)
        data.index.name = FC.TURBINE

        return data

    def reduce_turbines(self, vars_op):
        """
        Reduces the turbine dimension by some operation

        Parameters
        ----------
        vars_op: dict
            The operation per variable. Key: str, the variable
            name. Value: str, the operation, choices
            are: weights, mean_no_weights, sum, min, max.

        Returns
        -------
        data: pandas.DataFrame
            The results per state

        """
        states = self.results.coords[FC.STATE].to_numpy()

        rdata = {}
        for v, op in vars_op.items():
            vdata = self.results[v].to_numpy()
            nns = np.sum(np.isnan(vdata))
            assert nns == 0, (
                f"Found {nns} nan values for variable '{v}' of shape {vdata.shape}"
            )

            if op == "weights":
                rdata[v] = self.weinsum("s", vdata)
            elif op == "mean_no_weights":
                rdata[v] = np.mean(vdata, axis=1)
            elif op == "sum":
                rdata[v] = np.sum(vdata, axis=1)
            elif op == "min":
                rdata[v] = np.min(vdata, axis=1)
            elif op == "max":
                rdata[v] = np.max(vdata, axis=1)
            else:
                raise KeyError(
                    f"Unknown operation '{op}' for variable '{v}'. Please choose: weights, mean_no_weights, sum, min, max"
                )

        data = pd.DataFrame(index=states, data=rdata)
        data.index.name = FC.STATE

        return data

    def reduce_all(self, states_op, turbines_op):
        """
        Reduces states and turbine dimension by some operation

        Parameters
        ----------
        states_op: dict
            The states contraction operations.
            Key: str, the variable name. Value:
            str, the operation, choices are:
            sum, mean, min, max.
        turbines_op: dict
            The turbines contraction operations.
            Key: str, the variable name. Value:
            str, the operation, choices are:
            sum, mean, min, max.

        Returns
        -------
        data: dict
            The fully contracted results

        """
        sdata = self.reduce_states(states_op)

        rdata = {}
        for v, op in turbines_op.items():
            vdata = sdata[v].to_numpy()
            nns = np.sum(np.isnan(vdata))
            assert nns == 0, (
                f"Found {nns} nan values for variable '{v}' of shape {vdata.shape}"
            )

            if op == "weights":
                if states_op[v] == "weights":
                    rdata[v] = self.weinsum("", v)
                else:
                    rdata[v] = self.weinsum("", vdata[None, :])
            elif op == "mean_no_weights":
                rdata[v] = np.sum(vdata)
            elif op == "sum":
                rdata[v] = np.sum(vdata)
            elif op == "min":
                rdata[v] = np.min(vdata)
            elif op == "max":
                rdata[v] = np.max(vdata)
            else:
                raise KeyError(
                    f"Unknown operation '{op}' for variable '{v}'. Please choose: sum, mean, min, max, weights"
                )

        return rdata

    def calc_states_mean(self, vars, use_weights=True):
        """
        Calculates the mean wrt states.

        Parameters
        ----------
        vars: list of str
            The variables
        use_weights: bool
            Flag for using states weights for the mean

        Returns
        -------
        data: pandas.DataFrame
            The results per turbine

        """
        r = "weights" if use_weights else "mean_no_weights"
        if isinstance(vars, str):
            return self.reduce_states({vars: r})
        return self.reduce_states({v: r for v in vars})

    def calc_states_sum(self, vars):
        """
        Calculates the sum wrt states.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: pandas.DataFrame
            The results per turbine

        """
        return self.reduce_states({v: "sum" for v in vars})

    def calc_states_std(self, vars):
        """
        Calculates the standard deviation wrt states.

        Args:
            vars (_type_): _description_

        Returns:
            _type_: _description_
        """

        return self.reduce_states({v: "std" for v in vars})

    def calc_turbine_mean(self, vars):
        """
        Calculates the mean wrt turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: pandas.DataFrame
            The results per state

        """
        return self.reduce_turbines({v: "mean_no_weights" for v in vars})

    def calc_turbine_sum(self, vars):
        """
        Calculates the sum wrt turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: pandas.DataFrame
            The results per state

        """
        return self.reduce_turbines({v: "sum" for v in vars})

    def calc_farm_mean(self, vars):
        """
        Calculates the mean over states and turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: dict
            The fully contracted results

        """
        op = {v: "weights" for v in vars}
        return self.reduce_all(states_op=op, turbines_op=op)

    def calc_farm_sum(self, vars):
        """
        Calculates the sum over states and turbines.

        Parameters
        ----------
        vars: list of str
            The variables

        Returns
        -------
        data: dict
            The fully contracted results

        """
        op = {v: "sum" for v in vars}
        return self.reduce_all(states_op=op, turbines_op=op)

    def calc_mean_farm_power(self, ambient=False):
        """
        Calculates the mean total farm power.

        Parameters
        ----------
        ambient: bool
            Flag for ambient power

        Returns
        -------
        data: float
            The mean wind farm power

        """
        v = FV.P if not ambient else FV.AMB_P
        cdata = self.reduce_all(states_op={v: "weights"}, turbines_op={v: "sum"})
        return cdata[v]

    def calc_turbine_yield(
        self,
        algo=None,
        annual=False,
        ambient=False,
        hours=None,
        delta_t=None,
        P_unit_W=None,
    ):
        """
        Calculates the yield per turbine

        Parameters
        ----------
        algo: foxes.core.Algorithm, optional
            The algorithm, for P_nominal lookup
        annual: bool, optional
            Flag for returning annual results, by default False
        ambient: bool, optional
            Flag for ambient power, by default False
        hours: int, optional
            The duration time in hours, if not timeseries states
        delta_t: np.datetime64, optional
            The time delta step in case of time series data,
            by default automatically determined
        P_unit_W: float
            The power unit in Watts, 1000 for kW. Looked up
            in algorithm if not given

        Returns
        -------
        pandas.DataFrame
            A dataframe of yield values by turbine in GWh

        """
        if ambient:
            var_in = FV.AMB_P
            var_out = FV.AMB_YLD
        else:
            var_in = FV.P
            var_out = FV.YLD

        if algo is not None and P_unit_W is None:
            P_unit_W = np.array(
                [FC.P_UNITS[t.P_unit] for t in algo.farm_controller.turbine_types],
                dtype=config.dtype_double,
            )[:, None]
        elif algo is None and P_unit_W is not None:
            pass
        else:
            raise KeyError("Expecting either 'algo' or 'P_unit_W'")

        # compute yield per turbine
        if np.issubdtype(self.results[FC.STATE].dtype, np.datetime64):
            if hours is not None:
                raise KeyError("Unexpected parameter 'hours' for timeseries data")
            times = self.results[FC.STATE].to_numpy()
            if delta_t is None:
                delta_t = times[-1] - times[-2]
            duration = times[-1] - times[0] + delta_t
            duration_seconds = np.int64(duration.astype(np.int64) / 1e9)
            duration_hours = duration_seconds / 3600
        elif hours is None and annual:
            duration_hours = 8760
        elif hours is None:
            raise ValueError(
                "Expecting parameter 'hours' for non-timeseries data, or 'annual=True'"
            )
        else:
            duration_hours = hours
        yld = self.calc_states_mean(var_in) * duration_hours * P_unit_W / 1e9

        if annual:
            # convert to annual values
            yld *= 8760 / duration_hours

        yld.rename(columns={var_in: var_out}, inplace=True)
        return yld

    def add_capacity(self, algo=None, P_nom=None, ambient=False, verbosity=1):
        """
        Adds capacity to the farm results

        Parameters
        ----------
        algo: foxes.core.Algorithm, optional
            The algorithm, for nominal power calculation
        P_nom: list of float, optional
            Nominal power values for each turbine, if algo not given
        ambient: bool, optional
            Flag for calculating ambient capacity, by default False
        verbosity: int
            The verbosity level, 0 = silent

        """
        if ambient:
            var_in = FV.AMB_P
            var_out = FV.AMB_CAP
        else:
            var_in = FV.P
            var_out = FV.CAP

        # get results data for the vars variable (by state and turbine)
        vdata = self.results[var_in]

        if algo is not None and P_nom is None:
            P_nom = np.array(
                [t.P_nominal for t in algo.farm_controller.turbine_types],
                dtype=config.dtype_double,
            )
        elif algo is None and P_nom is not None:
            P_nom = np.array(P_nom, dtype=config.dtype_double)
        else:
            raise KeyError("Expecting either 'algo' or 'P_nom'")

        # add to farm results
        self.results[var_out] = vdata / P_nom[None, :]
        if verbosity > 0:
            if ambient:
                print("Ambient capacity added to farm results")
            else:
                print("Capacity added to farm results")

    def calc_farm_yield(self, turbine_yield=None, power_uncert=None, **kwargs):
        """
        Calculates yield, P75 and P90 at the farm level

        Parameters
        ----------
        turbine_yield: pandas.DataFrame, optional
            Yield values by turbine
        power_uncert: float, optional
            Uncertainty in the power value. Triggers
            P75 and P90 outputs
        kwargs: dict, optional
            Parameters for calc_turbine_yield(). Apply if
            turbine_yield is not given

        Returns
        -------
        farm_yield: float
            Farm yield result, same unit as turbine yield
        P75: float, optional
            The P75 value, same unit as turbine yield
        P90: float, optional
            The P90 value, same unit as turbine yield

        """
        if turbine_yield is None:
            yargs = dict(annual=True)
            yargs.update(kwargs)
            turbine_yield = self.calc_turbine_yield(**yargs)
        farm_yield = turbine_yield.sum()

        if power_uncert is not None:
            P75 = farm_yield * (1.0 - (0.675 * power_uncert))
            P90 = farm_yield * (1.0 - (1.282 * power_uncert))
            return farm_yield["YLD"], P75["YLD"], P90["YLD"]

        return farm_yield["YLD"]

    def add_efficiency(self, verbosity=1):
        """
        Adds efficiency to the farm results

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        P = self.results[FV.P].to_numpy()
        P0 = np.maximum(self.results[FV.AMB_P].to_numpy(), 1e-12)
        eff = np.minimum(P / P0, 1)
        eff[P < 1e-10] = 0
        self.results[FV.EFF] = (self.results[FV.AMB_P].dims, eff)
        if verbosity > 0:
            print("Efficiency added to farm results")

    def calc_farm_efficiency(self):
        """
        Calculates farm efficiency

        Returns
        -------
        eff: float
            The farm efficiency

        """
        P = self.calc_mean_farm_power()
        P0 = np.maximum(self.calc_mean_farm_power(ambient=True), 1e-14)
        return np.minimum(P / P0, 1)

    def gen_stdata(
        self,
        turbines,
        variable,
        fig=None,
        ax=None,
        figsize=None,
        legloc="lower right",
        animated=True,
        ret_im=True,
    ):
        """
        Generates state-turbine data,
        intended to be used in animations

        Parameters
        ----------
        turbines: list of int
            The turbines for which to scatter data
        variable: str
            The variable name
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        figsize: tuple, optional
            The figsize for plt.Figure
        legloc: str
            The legend location
        animated: bool
            Flag for animated output
        ret_im: bool
            Flag for image return,

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        im: List of matplotlib.collections.PathCollection, optional
            The scatter artists

        """

        if fig is None:
            hfig = plt.figure(figsize=figsize)
        else:
            hfig = fig
        if ax is None:
            hax = hfig.add_subplot(111)
        else:
            hax = ax

        hax.set_xlabel("State")
        hax.set_ylabel(variable)
        cc = cycler(color="bgrcmyk")

        data = self.results[variable].to_numpy()
        hasl = set()
        for si in range(len(data)):
            im = []
            hax.set_prop_cycle(cc)
            for ti in turbines:
                lbl = None if ti in hasl else f"Turbine {ti}"
                im += hax.plot(range(si), data[:si, ti], label=lbl, animated=animated)
                hasl.add(ti)

            hax.legend(loc=legloc)

            if ret_im:
                yield hfig, im
            else:
                yield hfig



import numpy as np
import matplotlib.pyplot as plt
from xarray import Dataset
from matplotlib.projections.polar import PolarAxes
from matplotlib.lines import Line2D

from foxes.algorithms import Downwind
from foxes.core import WindFarm, Turbine
from foxes.models import ModelBook
import foxes.variables as FV
import foxes.constants as FC

from .output import Output


class RosePlotOutput(Output):
    """
     Class for rose plot creation

     Attributes
     ----------
     results: pandas.DataFrame
         The calculation results (farm or points)

    :group: output

    """

    def __init__(
        self,
        farm_results=None,
        point_results=None,
        use_points=False,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset, optional
            The farm results
        point_results: xarray.Dataset, optional
            The point results
        use_points: bool
            Flag for using points in cases where both
            farm and point results are given
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        if use_points or (farm_results is None and point_results is not None):
            self.results = point_results
            self._rtype = FC.POINT
        elif farm_results is not None:
            self.results = farm_results
            self._rtype = FC.TURBINE
        else:
            raise KeyError("Require either farm_results or point_results")

    @classmethod
    def get_data_info(cls, dname):
        """
        Returns default description for a variable.

        Parameters
        ----------
        dname: str
            The variable name

        Returns
        -------
        title: str
            The long name of the variable
        legend: str
            The legend/axis text

        """

        if dname == FV.D:
            return "Rotor diameter", f"{FV.D} [m]"
        if dname == FV.H:
            return "Hub height", f"{FV.H} [m]"

        if dname == FV.WS:
            return "Wind speed", f"{FV.WS} [m/s]"
        if dname == FV.REWS:
            return "Rotor equivalent wind speed", f"{FV.REWS} [m/s]"
        if dname == FV.REWS2:
            return "Rotor equivalent wind speed (2nd moment)", f"{FV.REWS2} [m/s]"
        if dname == FV.REWS3:
            return "Rotor equivalent wind speed (3rd moment)", f"{FV.REWS3} [m/s]"
        if dname == FV.WD:
            return "Wind direction", f"{FV.WD} [deg]"
        if dname == FV.TI:
            return "Turbulence intensity", f"{FV.TI} [1]"
        if dname == FV.RHO:
            return "Air density", f"{FV.RHO} [kg/m3]"
        if dname == FV.CT:
            return "Thrust coefficient", f"{FV.CT} [1]"
        if dname == FV.P:
            return "Power", f"{FV.P} [kW]"
        if dname == FV.YAW:
            return "Yaw angle", f"{FV.YAW} [deg]"
        if dname == FV.YAWM:
            return "Yaw misalignment", f"{FV.YAWM} [deg]"

        if dname in FV.amb2var:
            title, legend = cls.get_data_info(FV.amb2var[dname])
            return f"Ambient {title.lower()}", f"AMB_{legend}"

        return dname, dname

    def get_data(
        self,
        wd_sectors,
        ws_var,
        ws_bins,
        wd_var=FV.AMB_WD,
        turbine=0,
        point=0,
        add_inf=False,
    ):
        """
        Generates the plot data

        Parameters
        ----------
        wd_sectors: int
            The number of wind rose sectors
        ws_var: str
            The wind speed variable
        ws_bins: list of float
            The wind speed bins
        wd_var: str
            The wind direction variable
        turbine: int
            The turbine index, for weights and for
            data if farm_results are given
        point: int
            The point index, for data if point_results
            are given
        add_inf: bool
            Add an upper bin up to infinity

        Returns
        -------
        data: xarray.Dataset
            The plot data

        """
        if self.results[FV.WEIGHT].dims == (FC.STATE,):
            w = self.results[FV.WEIGHT].to_numpy()
        elif self.results[FV.WEIGHT].dims == (FC.STATE, FC.TURBINE):
            w = self.results[FV.WEIGHT].to_numpy()[:, turbine]
        elif self.results[FV.WEIGHT].dims == (FC.STATE, FC.POINT):
            w = self.results[FV.WEIGHT].to_numpy()[:, point]
        else:
            raise ValueError(
                f"Wrong dimensions for '{FV.WEIGHT}'. Expecting {(FC.STATE,)}, {(FC.STATE, FC.TURBINE)} or {(FC.STATE, FC.POINT)}, got {self.results[FV.WEIGHT].dims}"
            )

        if add_inf:
            ws_bins = list(ws_bins) + [np.inf]
        t = turbine if self._rtype == FC.TURBINE else point
        ws = self.results[ws_var].to_numpy()[:, t]
        wd = self.results[wd_var].to_numpy()[:, t].copy()
        wd_delta = 360 / wd_sectors
        wd[wd >= 360 - wd_delta / 2] -= 360
        wd_bins = np.arange(-wd_delta / 2, 360, wd_delta)
        ws_bins = np.asarray(ws_bins, dtype=ws.dtype)

        freq = 100 * np.histogram2d(wd, ws, (wd_bins, ws_bins), weights=w)[0]

        data = Dataset(
            coords={
                wd_var: np.arange(0, 360, wd_delta),
                ws_var: 0.5 * (ws_bins[:-1] + ws_bins[1:]),
            },
            data_vars={
                f"bin_min_{wd_var}": (wd_var, wd_bins[:-1]),
                f"bin_max_{wd_var}": (wd_var, wd_bins[1:]),
                f"bin_min_{ws_var}": (ws_var, ws_bins[:-1]),
                f"bin_max_{ws_var}": (ws_var, ws_bins[1:]),
                "frequency": ((wd_var, ws_var), freq),
            },
            attrs={
                f"{wd_var}_bounds": wd_bins,
                f"{ws_var}_bounds": ws_bins,
            },
        )

        return data

    def get_figure(
        self,
        wd_sectors,
        ws_var,
        ws_bins,
        wd_var=FV.AMB_WD,
        fig=None,
        ax=None,
        figsize=None,
        freq_delta=3,
        cmap="summer",
        title=None,
        legend_pars=None,
        ret_data=False,
        **kwargs,
    ):
        """
        Creates the figure

        Parameters
        ----------
        wd_sectors: int
            The number of wind rose sectors
        ws_var: str
            The wind speed variable
        ws_bins: list of float
            The wind speed bins
        wd_var: str
            The wind direction variable
        fig: pyplot.Figure, optional
            The figure object
        ax: pyplot.Axes, optional
            The axes object
        figsize: tuple, optional
            The figsize argument for plt.subplots
        freq_delta: int
            The frequency delta for the label
            in percent
        cmap: str
            The color map
        title: str, optional
            The title
        legend_pars: dict, optional
            Parameters for the legend
        ret_data: bool
            Flag for returning wind rose data
        kwargs: dict, optional
            Additional parameters for get_data

        Returns
        -------
        ax: pyplot.Axes
            The axes object
        data: xarray.Dataset, optional
            The plot data

        """
        if self.nofig:
            return None

        data = self.get_data(wd_sectors, ws_var, ws_bins, wd_var, **kwargs)

        n_wsb = data.sizes[ws_var]
        n_wdb = data.sizes[wd_var]
        ws_bins = np.asarray(data.attrs[f"{ws_var}_bounds"])
        wd_cent = np.mod(90 - data[wd_var].to_numpy(), 360)
        wd_cent = np.radians(wd_cent)
        wd_delta = 360 / n_wdb
        wd_width = np.radians(0.9 * wd_delta)
        freq = data["frequency"].to_numpy()

        if ax is not None:
            if not isinstance(ax, PolarAxes):
                raise TypeError(
                    f"Require axes of type '{PolarAxes.__name__}' for '{type(self).__name__}', got '{type(ax).__name__}'"
                )
        else:
            fig, ax = plt.subplots(figsize=figsize, subplot_kw={"projection": "polar"})

        bcmap = plt.get_cmap(cmap, n_wsb)
        color_list = bcmap(np.linspace(0, 1, n_wsb))

        bottom = np.zeros(n_wdb)
        for wsi in range(n_wsb):
            ax.bar(
                wd_cent,
                freq[:, wsi],
                bottom=bottom,
                width=wd_width,
                color=color_list[wsi],
            )
            bottom += freq[:, wsi]

        fmax = np.max(np.sum(freq, axis=1))
        freq_delta = int(freq_delta)
        freq_ticks = np.arange(0, fmax + freq_delta / 2, freq_delta, dtype=np.int32)[1:]

        tksl = np.arange(0, 360, max(wd_delta, 30))
        tks = np.radians(np.mod(90 - tksl, 360))
        ax.set_xticks(tks, [f"{int(d)}°" for d in tksl])
        ax.set_yticks(freq_ticks, [f"{f}%" for f in freq_ticks])
        ax.set_title(title)

        llines = [Line2D([0], [0], color=c, lw=10) for c in np.flip(color_list, axis=0)]
        lleg = [
            f"[{ws_bins[i]:.1f}, {ws_bins[i + 1]:.1f})"
            for i in range(n_wsb - 1, -1, -1)
        ]
        lpars = dict(
            loc="upper left",
            bbox_to_anchor=(0.8, 0.5),
            title=f"{ws_var}",
        )
        wsl = [FV.WS, FV.REWS, FV.REWS2, FV.REWS3]
        wsl += [FV.var2amb[v] for v in wsl]
        if ws_var in wsl:
            lpars["title"] += " [m/s]"
        if legend_pars is not None:
            lpars.update(legend_pars)
        ax.legend(llines, lleg, **lpars)

        if ret_data:
            return ax, data
        else:
            return ax

    def write_figure(self, file_name, *args, ret_data=False, **kwargs):
        """
        Write rose plot to file

        Parameters
        ----------
        file_name: str
            Name of the output file
        args: tuple, optional
            Additional parameters for get_figure
        ret_data: bool
            Flag for returning wind rose data
        kwargs: dict, optional
            Additional parameters for get_figure

        Returns
        -------
        data: pd.DataFrame, optional
            The wind rose data

        """
        if self.nofig:
            return None

        r = self.get_figure(*args, ret_data=ret_data, **kwargs)
        fpath = self.get_fpath(file_name)
        if ret_data:
            r[0].get_figure().savefig(fpath, bbox_inches="tight")
            return r[1]
        else:
            r.get_figure().savefig(fpath, bbox_inches="tight")


class StatesRosePlotOutput(RosePlotOutput):
    """
    Class for rose plot creation directly from states
    :group: output
    """

    def __init__(
        self,
        states,
        point,
        mbook=None,
        ws_var=FV.AMB_REWS,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        states: foxes.core.States
            The states from which to compute the wind rose
        point: numpy.ndarray
            The evaluation point, shape: (3,)
        mbook: foxes.models.ModelBook, optional
            The model book
        ws_var: str
            The wind speed variable name
        kwargs: dict, optional
            Additional parameters for the base class

        """
        farm = WindFarm()
        farm.add_turbine(
            Turbine(
                xy=point[:2],
                H=point[2],
                turbine_models=["null_type"],
            ),
            verbosity=0,
        )

        mbook = mbook if mbook is not None else ModelBook()
        algo = Downwind(farm, states, wake_models=[], mbook=mbook, verbosity=0)

        results = algo.calc_farm(ambient=True).rename_vars({ws_var: FV.AMB_WS})

        super().__init__(results, **kwargs)


class WindRoseBinPlot(Output):
    """
    Plots mean data in wind rose bins

    Attributes
    ----------
    farm_results: xarray.Dataset
        The wind farm results

    :group: output

    """

    def __init__(self, farm_results, **kwargs):
        """
        Constructor

        Parameters
        ----------
        farm_results: xarray.Dataset
            The wind farm results
        kwargs: dict, optional
            Parameters for the base class

        """
        super().__init__(**kwargs)
        self.farm_results = farm_results

    def get_data(
        self,
        variable,
        ws_bins,
        wd_sectors=12,
        wd_var=FV.AMB_WD,
        ws_var=FV.AMB_REWS,
        turbine=0,
        contraction="weights",
    ):
        """
        Generates the plot data

        Parameters
        ----------
        variable: str
            The variable name
        ws_bins: list of float
            The wind speed bins
        wd_var: str
            The wind direction variable
        ws_var: str
            The wind speed variable
        turbine: int
            The turbine index
        contraction: str
            The contraction method for states:
            weights, mean_no_weights, sum_no_weights

        Returns
        -------
        data: xarray.Dataset
            The plot data

        """
        if self.farm_results[FV.WEIGHT].dims == (FC.STATE,):
            w = self.farm_results[FV.WEIGHT].to_numpy()
        elif self.farm_results[FV.WEIGHT].dims == (FC.STATE, FC.TURBINE):
            w = self.farm_results[FV.WEIGHT].to_numpy()[:, turbine]
        else:
            raise ValueError(
                f"Wrong dimensions for '{FV.WEIGHT}'. Expecting {(FC.STATE,)} or {(FC.STATE, FC.TURBINE)}, got {self.farm_results[FV.WEIGHT].dims}"
            )

        var = self.farm_results[variable].to_numpy()[:, turbine]
        ws = self.farm_results[ws_var].to_numpy()[:, turbine]
        wd = self.farm_results[wd_var].to_numpy()[:, turbine].copy()
        wd_delta = 360 / wd_sectors
        wd[wd >= 360 - wd_delta / 2] -= 360
        wd_bins = np.arange(-wd_delta / 2, 360, wd_delta)
        ws_bins = np.asarray(ws_bins)

        if contraction == "weights":
            z = np.histogram2d(wd, ws, (wd_bins, ws_bins), weights=w)[0]
            z[z < 1e-13] = np.nan
            z = np.histogram2d(wd, ws, (wd_bins, ws_bins), weights=w * var)[0] / z
        elif contraction == "mean_no_weights":
            z = np.histogram2d(wd, ws, (wd_bins, ws_bins))[0].astype(w.dtype)
            z[z < 1] = np.nan
            z = np.histogram2d(wd, ws, (wd_bins, ws_bins), weights=var)[0] / z
        elif contraction == "sum_no_weights":
            z = np.histogram2d(wd, ws, (wd_bins, ws_bins), weights=var)[0]
        else:
            raise KeyError(
                f"Contraction '{contraction}' not supported. Choices: weights, mean_no_weights, sum_no_weights"
            )

        data = Dataset(
            coords={
                wd_var: 0.5 * (wd_bins[:-1] + wd_bins[1:]),
                ws_var: 0.5 * (ws_bins[:-1] + ws_bins[1:]),
            },
            data_vars={
                variable: ((wd_var, ws_var), z),
            },
            attrs={
                f"{wd_var}_bounds": wd_bins,
                f"{ws_var}_bounds": ws_bins,
            },
        )

        return data

    def get_figure(
        self,
        variable,
        ws_bins,
        wd_sectors=12,
        wd_var=FV.AMB_WD,
        ws_var=FV.AMB_REWS,
        turbine=0,
        contraction="weights",
        fig=None,
        ax=None,
        title=None,
        figsize=None,
        ret_data=False,
        **kwargs,
    ):
        """
        Creates the figure

        Parameters
        ----------
        variable: str
            The variable name
        ws_bins: list of float
            The wind speed bins
        wd_var: str
            The wind direction variable
        ws_var: str
            The wind speed variable
        turbine: int
            The turbine index
        contraction: str
            The contraction method for states:
            weights, mean_no_weights, sum_no_weights
        fig: pyplot.Figure, optional
            The figure object
        ax: pyplot.Axes, optional
            The axes object
        title: str, optional
            The title
        figsize: tuple, optional
            The figsize argument for plt.subplots
        ret_data: bool
            Flag for returning wind rose data
        kwargs: dict, optional
            Additional parameters for plt.pcolormesh

        Returns
        -------
        ax: pyplot.Axes
            The axes object

        """
        if self.nofig:
            return None

        data = self.get_data(
            variable=variable,
            ws_bins=ws_bins,
            wd_sectors=wd_sectors,
            wd_var=wd_var,
            ws_var=ws_var,
            turbine=turbine,
            contraction=contraction,
        )

        wd_delta = 360 / data.sizes[wd_var]
        wd_bins = np.mod(90 - data.attrs[f"{wd_var}_bounds"], 360)
        wd_bins = np.radians(wd_bins)
        ws_bins = data.attrs[f"{ws_var}_bounds"]

        if ax is not None:
            if not isinstance(ax, PolarAxes):
                raise TypeError(
                    f"Require axes of type '{PolarAxes.__name__}' for '{type(self).__name__}', got '{type(ax).__name__}'"
                )
        else:
            fig, ax = plt.subplots(figsize=figsize, subplot_kw={"projection": "polar"})

        y, x = np.meshgrid(ws_bins, wd_bins)
        z = data[variable].to_numpy()

        prgs = {"shading": "flat"}
        prgs.update(kwargs)

        img = ax.pcolormesh(x, y, z, **prgs)

        tksl = np.arange(0, 360, max(wd_delta, 30))
        tks = np.radians(np.mod(90 - tksl, 360))
        ax.set_xticks(tks, [f"{d}°" for d in tksl])
        ax.set_yticks(ws_bins)
        ax.set_title(title)
        cbar = fig.colorbar(img, ax=ax, pad=0.12)
        cbar.ax.set_title(variable)

        if ret_data:
            return ax, data
        else:
            return ax

    def write_figure(self, file_name, *args, ret_data=False, **kwargs):
        """
        Write rose plot to file

        Parameters
        ----------
        file_name: str
            Name of the output file
        args: tuple, optional
            Additional parameters for get_figure
        ret_data: bool
            Flag for returning wind rose data
        kwargs: dict, optional
            Additional parameters for get_figure

        Returns
        -------
        data: pd.DataFrame, optional
            The wind rose data

        """
        if self.nofig:
            return None

        r = self.get_figure(*args, ret_data=ret_data, **kwargs)
        fpath = self.get_fpath(file_name)
        if ret_data:
            r[0].get_figure().savefig(fpath, bbox_inches="tight")
            return r[1]
        else:
            r.get_figure().savefig(fpath, bbox_inches="tight")
